from test.unittests.commands.xml_validate.fixtures_data_deserialised import *  # noqa: F403
from test.unittests.commands.xml_validate.fixtures_data_rdf import *  # noqa: F403
from test.unittests.commands.xml_validate.fixtures_onto import *  # noqa: F403
from test.unittests.commands.xml_validate.fixtures_xml_data import *  # noqa: F403
